@@ -49,8 +49,8 @@
 #include <dpkg/arch.h>
 #include <dpkg/subproc.h>
 #include <dpkg/command.h>
-#include <dpkg/myopt.h>
 #include <dpkg/pkg-spec.h>
+#include <dpkg/options.h>
 
 #include "main.h"
 #include "filesdb.h"
@@ -168,8 +168,7 @@ usage(const struct cmdinfo *ci, const char *value)
   exit(0);
 }
 
-const char thisname[]= "dpkg";
-const char printforhelp[]= N_(
+static const char printforhelp[] = N_(
 "Type dpkg --help for help about installing and deinstalling packages [*];\n"
 "Use `dselect' or `aptitude' for user-friendly package management;\n"
 "Type dpkg -Dhelp for a list of dpkg debug flag values;\n"
@@ -189,6 +188,7 @@ int fc_conff_old=0, fc_conff_def=0;
 int fc_conff_ask = 0;
 int fc_unsafe_io = 0;
 int fc_badverify = 0;
+int fc_badversion = 0;
 
 int errabort = 50;
 static const char *admindir = ADMINDIR;
@@ -231,6 +231,8 @@ static const struct forceinfo {
     ' ', N_("PATH is missing important programs, problems likely") },
   { "bad-verify",          &fc_badverify,
     ' ', N_("Install a package even if it fails authenticity check") },
+  { "bad-version",         &fc_badversion,
+    ' ', N_("Process even packages with wrong versions") },
   { "overwrite",           &fc_overwrite,
     ' ', N_("Overwrite a file from one package with another") },
   { "overwrite-diverted",  &fc_overwritediverted,
@@ -540,15 +542,15 @@ static void setforce(const struct cmdinfo *cip, const char *value) {
   }
 }
 
-void execbackend(const char *const *argv) DPKG_ATTR_NORET;
-void commandfd(const char *const *argv);
+int execbackend(const char *const *argv) DPKG_ATTR_NORET;
+int commandfd(const char *const *argv);
 
 /* This table has both the action entries in it and the normal options.
  * The action entries are made with the ACTION macro, as they all
  * have a very similar structure. */
 static const struct cmdinfo cmdinfos[]= {
 #define ACTIONBACKEND(longopt, shortopt, backend) \
- { longopt, shortopt, 0, NULL, NULL, setaction, 0, (void *)backend, (void_func *)execbackend }
+ { longopt, shortopt, 0, NULL, NULL, setaction, 0, (void *)backend, execbackend }
 
   ACTION( "install",                        'i', act_install,              archivefiles    ),
   ACTION( "unpack",                          0,  act_unpack,               archivefiles    ),
@@ -625,14 +627,14 @@ static const struct cmdinfo cmdinfos[]= {
   ACTIONBACKEND( "info",		'I', BACKEND),
   ACTIONBACKEND( "field",		'f', BACKEND),
   ACTIONBACKEND( "extract",		'x', BACKEND),
-  ACTIONBACKEND( "new",			0,  BACKEND),
-  ACTIONBACKEND( "old",			0,  BACKEND),
   ACTIONBACKEND( "vextract",		'X', BACKEND),
   ACTIONBACKEND( "fsys-tarfile",	0,   BACKEND),
   { NULL,                0,   0, NULL,          NULL,      NULL,          0 }
 };
 
-void execbackend(const char *const *argv) {
+int
+execbackend(const char *const *argv)
+{
   struct command cmd;
   char *arg;
 
@@ -650,16 +652,18 @@ void execbackend(const char *const *argv) {
   command_exec(&cmd);
 }
 
-void commandfd(const char *const *argv) {
+int
+commandfd(const char *const *argv)
+{
   struct varbuf linevb = VARBUF_INIT;
   const char * pipein;
   const char **newargs = NULL;
   char *ptr, *endptr;
   FILE *in;
   unsigned long infd;
+  int ret = 0;
   int c, lno, i;
   bool skipchar;
-  void (*actionfunction)(const char *const *argv);
 
   pipein = *argv++;
   if (pipein == NULL)
@@ -738,26 +742,29 @@ void commandfd(const char *const *argv) {
         newargs[i] = m_strdup(newargs[i]);
 
     setaction(NULL, NULL);
-    myopt((const char *const**)&newargs,cmdinfos);
+    myopt((const char *const **)&newargs, cmdinfos, printforhelp);
     if (!cipaction) badusage(_("need an action option"));
 
-    actionfunction = (void (*)(const char *const *))cipaction->arg_func;
-    actionfunction(newargs);
+    ret |= cipaction->action(newargs);
 
     pop_error_context(ehflag_normaltidy);
   }
+
+  return ret;
 }
 
 int main(int argc, const char *const *argv) {
-  void (*actionfunction)(const char *const *argv);
+  int ret;
 
-  setlocale(LC_ALL, "");
+  if (getenv("DPKG_UNTRANSLATED_MESSAGES") == NULL)
+    setlocale(LC_ALL, "");
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
 
+  dpkg_set_progname("dpkg");
   standard_startup();
   loadcfgfile(DPKG, cmdinfos);
-  myopt(&argv, cmdinfos);
+  myopt(&argv, cmdinfos, printforhelp);
 
   if (!cipaction) badusage(_("need an action option"));
 
@@ -779,14 +786,12 @@ int main(int argc, const char *const *argv) {
 
   filesdbinit();
 
-  actionfunction = (void (*)(const char *const *))cipaction->arg_func;
-
-  actionfunction(argv);
+  ret = cipaction->action(argv);
 
   if (is_invoke_action(cipaction->arg_int))
     run_invoke_hooks(cipaction->olong, post_invoke_hooks);
 
   standard_shutdown();
 
-  return reportbroken_retexitstatus();
+  return reportbroken_retexitstatus(ret);
 }
